/*
 * fwupload.c
 * Copyright (c) 2007 Hannu Nuotio <hannu.nuotio@pp.inet.fi>
 * Written by Hannu Nuotio <hannu.nuotio@pp.inet.fi>
 */

/*
 * fwupload upgrades the 1541-III firmware.
 * Use ihex2boot to convert Intel HEX firmware files to the
 * "boot" format used by this program.
 * The firmware can be read from a file or stdin.
 *
 * usage:
 *  fwupload [-hqvsc] [-t device] [firmwarefile]
 *	-h		help
 *	-q		quiet
 *	-v		verbose
 *	-c		disable block check (DANGEROUS)
 *	-t device	serial device (default = /dev/ttyS0)
 *	firmwarefile	firmware in "boot" format
 */

#define RX_TIMEOUT 20	// timeout for serial read (deciseconds)
#define MAX_CRC_RETRIES 20	// max. retries if CRC error

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>

void usage()
{
fprintf(stderr,"fwupload v.1.0\n");
fprintf(stderr,"--------------\n");
fprintf(stderr," fwupload [-hqvsc] [-t device] [firmwarefile]\n\n");
fprintf(stderr,"Upgrades the 1541-III firmware.\n");
fprintf(stderr,"Uses stdin if no filename is given.\n\n");
fprintf(stderr," -h		this help text\n");
fprintf(stderr," -q		quiet\n");
fprintf(stderr," -v		verbose\n");
fprintf(stderr," -c		disable block check (DANGEROUS)\n");
fprintf(stderr," -t device	serial device (default = /dev/ttyS0)\n");
fprintf(stderr," firmwarefile	firmware in \"boot\" format\n");
}

int main(int argc, char *argv[])
{
int com, c, res, vflg=1, len, ocnt, state, cflg=1, i, retries=0, resends=0;
FILE *ifile;
struct termios oldtio, newtio;
unsigned char buf[69], x, ocrc;

char *ifilename;
char *comdevice = "/dev/ttyS0";
extern char *optarg;
extern int optind, optopt;

while((c = getopt(argc,argv,"hqvct:")) != -1) {
	switch(c) {
	case 'h':
		usage();
		return 0;
		break;	
	case 'q':
		vflg=0;
		break;
	case 'v':
		vflg=2;
		break;
	case 'c':
		cflg=0;
		break;
	case 't':
		comdevice=optarg;
		break;
	case ':':
		fprintf(stderr,"-%c requires an operand\n",c);
		return 1;
		break;
	case '?':
		fprintf(stderr,"-%c is unknown\n",c);
		return 1;
		break;
	}
}

if (vflg) {fprintf(stderr,"fwupload\n--------\n");}

if ((ifilename=argv[optind])==NULL) {
	fprintf(stderr,"No filename, using stdin.\n");
	ifile=stdin;
	ifilename="stdin";
} else {
	ifile=fopen(ifilename,"rb");
	if (ifile == NULL) {perror(ifilename); return 1;}
	if (vflg) {fprintf(stderr,"File %s found.\n",ifilename);}
}

if (vflg) {fprintf(stderr,"Transfering %s -> %s\n",ifilename,comdevice);}

com = open(comdevice, O_RDWR | O_NOCTTY ); 
if (com < 0) {perror(comdevice); return 1; }

if (vflg) {fprintf(stdout,"%s opened.\n",comdevice);}
        
tcgetattr(com,&oldtio);
        
memset(&newtio, 0, sizeof(newtio));
newtio.c_cflag = B115200 | CS8 | CLOCAL | CREAD;
newtio.c_iflag = IGNPAR;
newtio.c_oflag = 0;
        
newtio.c_lflag = 0;
         
newtio.c_cc[VTIME] = RX_TIMEOUT;
newtio.c_cc[VMIN]  = 0;

tcflush(com, TCIFLUSH);
tcsetattr(com,TCSANOW,&newtio);

if (vflg) {fprintf(stderr,"Sending start byte...");}
res = write(com,"\xc1",1);
if (res < 0) {perror(comdevice); return 1; }
if (vflg) {fprintf(stderr,"ok...");}

res = read(com,&x,1);
if (res == 0) {fprintf(stderr,"IdTypePIC not received!\n"); return 1;}
if (x != 0x4F) {fprintf(stderr,"incorrect IdTypePIC (%02x)\n",x); return 1;}
if (vflg) {fprintf(stderr,"IdTypePIC received. Proceeding to main loop...");}

ocnt=0;
state=1;

while(state) {
	res = read(com,&x,1);
	if (res == 0) {fprintf(stderr,"Ack not received!\n"); return 1;}
	if (x == 'N') {
		resends++;
		if (++retries<=MAX_CRC_RETRIES) {
			if (vflg==2) {fprintf(stderr,"CRC error, resend %i/%i...",retries,MAX_CRC_RETRIES);}
			res = write(com,buf,69);
			if (res < 0) {perror(comdevice); return 1; }
		} else {
			fprintf(stderr,"CRC error! (block %i)\n",ocnt); 
			return 1;
		}
	} else {
		if (x != 'K') {fprintf(stderr,"Invalid ack (%02x)!\n",x); return 1;}
		if (vflg==2) {fprintf(stderr,"ok.\n");}
		len=fread(buf,1,69,ifile);
		if (len < 0) {perror(ifilename); return 1;}
		if ((len > 0)&&(len <69)) {fprintf(stderr,"Error: read only %i bytes.\n",len); return 1;}
		if (len == 0) {
			state=0;
		} else {
			ocnt++;
			if (cflg) {
				if (vflg==2) {fprintf(stderr,"Checking block %i...",ocnt);}
				ocrc=0;
				for (i=0;i<69;ocrc+=buf[i++]);
				if (ocrc!=0) {fprintf(stderr,"CRC error (%i) on block %i!\n",ocrc,ocnt); return 1;}
				if ((buf[2]&0x3f)!=0) {fprintf(stderr,"mod 64 error on block %i!\n",ocnt); return 1;}
				if (buf[3]!=64) {fprintf(stderr,"len != 64 (%i) error on block %i!\n",buf[3],ocnt); return 1;}
			}
			if (vflg==2) {fprintf(stderr,"Sending block %i...",ocnt);}
			res = write(com,buf,69);
			if (res < 0) {perror(comdevice); return 1; }
			if (vflg==2) {fprintf(stderr,"sent...");}
			retries=0;
		}
	}
}

if (vflg) {fprintf(stderr,"OK! %i*69=%i bytes sent (%i resends). ",ocnt,ocnt*69,resends);}

if (vflg) {fprintf(stderr,"Cleaning up...");}
if(ifilename!=NULL) {
	res=fclose(ifile);
	if (res < 0) {perror(ifilename); return 1;}
}
tcsetattr(com,TCSANOW,&oldtio);
close(com);
if (vflg) {fprintf(stderr,"done.\n");}

return 0;
}
